// LCD-Routines for ASCII-Displays with HD44780U oder compatible controller
// Johann Gysin
//
// History
// 13.06.2002  Johann Gysin  Erste voll funktionsfhige Version
// 19.06.2002  Johann Gysin  Nur druckbare Zeichen werden auf Display ausgegeben (undruckbare unterdrckt)
// 27.06.2002  Johann Gysin  Untersttzung fr 4 Zeiliges Display mit 20 Zeichen eingebaut
// 02.07.2002  Johann Gysin  Busy-Flag-Abfrage statt Warteschleife; sorgt fr optimale Anzeigegeschwindigkeit
// 15.08.2002  Johann Gysin  Zeichenkonvertierung fr Umlaute eingebaut
// 21.10.2002  Johann Gysin  Portzuweisungen erfolgen per Define-Anweisung, Funktion ClearLine hinzugfgt
// 25.10.2002  Johann Gysin  Wartefunktion LCD_Wait_Time() an MSP430 mit 4MHz angepasst
// 01.07.2003  Johann Gysin  Wartefunktion deutlich verlngert wegen Aufstartproblemen
// 20.10.2003  Johann Gysin  Modifications for H8S/2329 Board

#include "LCD.h"
#include "ioh82329.h"

#define PDirReg PDDDR // Port Direction Register (Input/Output)
#define POutReg PDDR  // Port Output Register
#define PInpReg PORTD // Port Input Register

unsigned char DDRAM_Offset_line_1;
unsigned char DDRAM_Offset_line_2;
unsigned char DDRAM_Offset_line_3;
unsigned char DDRAM_Offset_line_4;
unsigned char AnzahlController;
unsigned char AnzahlZeichenProline;
unsigned char LCD_DisplayType;

void LCD_Wait_Time(int Time);
void LCD_Wait(unsigned char Enable);
void LCD_Init(unsigned char DisplayType);
void LCD_WriteInstr(unsigned char LCD_Controller, unsigned char Instr);
void LCD_Clear(void);
void LCD_Control(unsigned char LCD_Controller, unsigned char disp_on, unsigned char cursor_on, unsigned char blink);
char LCD_SetCursor(unsigned char line, unsigned char index);
void LCD_WriteChar(unsigned char LCD_Controller, unsigned char Data);
void LCD_WriteStr(unsigned char line, unsigned char index, const unsigned char *str);
void LCD_WriteInt(unsigned char line, unsigned char index, int zahl);
void LCD_ClearLine(unsigned char line);

void LCD_Wait_Time(int Time) // Time = Wartezeit in Millisekunden
  {
  volatile int a;
  volatile int b;

  for (a=0; a<Time; a++)
    {
    for (b=0; b<2000; b++) // Wartet ca. 1ms (muss fr 2329 noch angepasst werden...)
      {
      }
    }

  }
  
void LCD_Wait(unsigned char Enable)
  {
  unsigned char busy;
  
  PDirReg = B6+B5+B4;        // Alles ausser D4 bis D7 als Ausgang
  do
    {
    POutReg = B5;            // Read from Display
    POutReg = Enable | B5;   // Read und Enable des richtigen Controllers (High Nibble)
    busy = PInpReg & B3;     // Busy Flag lesen
    POutReg = B5;            // Enable weg
    POutReg = Enable | B5;   // Read und Enable (Low Nibble)
    POutReg = B5;            // Enable weg
    }
  while (busy);             // Wiederholen bis Busy-Flag vom LCD-Controller gelscht ist
  POutReg = 0;              // Disable / Write
  PDirReg = 255;            // Alle Pins als Ausgang
  }

void LCD_Init(unsigned char DisplayType)
  {
  unsigned char enable;
  
  LCD_DisplayType = DisplayType;
    
  enable = B4; // Beide LCD-Controller enablen
    
  PDirReg = 255;   // Pord D als Ausgang
    
  LCD_Wait_Time(100);  // min 15ms Warten
  POutReg = enable;      // Enable1, Enable2
  POutReg = enable+B0+B1; // Enable1, Enable2, Data (3) // Init
  POutReg = B0+B1;          // Data (3)
  
  LCD_Wait_Time(25);  // min 5ms Warten
  
  POutReg = enable;      // Enable1, Enable2
  POutReg = enable+B0+B1; // Enable1, Enable2, Data (3) // Init
  POutReg = B0+B1;          // Data (3)
  LCD_Wait_Time(5);   // min 160us Warten
  
  POutReg = enable;      // Enable1, Enable2
  POutReg = enable+B0+B1; // Enable1, Enable2, Data (3) // Init
  POutReg = B0+B1;          // Data (3)
  LCD_Wait_Time(2);   // min 160us Warten
  POutReg = enable;      // Enable1, Enable2
  POutReg = enable+B1; // Enable1, Enable2, Data (2)    // 4-Bit Mode
  POutReg = B1;          // Data (2)
  LCD_Wait_Time(2);   // min 160us Warten
   
  // From now, 2 Nibbles required
  if (DisplayType == 3)
    {
    LCD_WriteInstr(3, 9);       // 4-Line Mode
    }
    
  LCD_WriteInstr(3, B3+B5);   // DataLength 4-Bit, 2 Lines, 5x7Dots Font
  
  LCD_Clear();                // Clear Display
  
  LCD_Control(3, 1, 0, 0);    // Display On, Cursor Off, Blink Off
  LCD_WriteInstr(3, B1);      // Set DDRAM Addr. 0

  LCD_Wait_Time(1);           // Nochmals warten zur Sicherheit
  
  // Variabeln setzen fr verschiedene Display-Typen
  switch (DisplayType)
    {
    // Display mit 4 linen  40 Zeichen
    case 0: DDRAM_Offset_line_1 = 0;
            DDRAM_Offset_line_2 = 64;
            DDRAM_Offset_line_3 = 0;
            DDRAM_Offset_line_4 = 64;
            AnzahlController = 2;
            AnzahlZeichenProline = 40;
            break;
    
    // Display mit 4 linen  20 Zeichen
    case 1: DDRAM_Offset_line_1 = 0;
            DDRAM_Offset_line_2 = 64;
            DDRAM_Offset_line_3 = 20;
            DDRAM_Offset_line_4 = 84;
            AnzahlController = 1;
            AnzahlZeichenProline = 20;
            break; 
    
    // Display mit 2 linen  24 Zeichen
    case 2: DDRAM_Offset_line_1 = 0;
            DDRAM_Offset_line_2 = 64;
            DDRAM_Offset_line_3 = 20;
            DDRAM_Offset_line_4 = 84;
            AnzahlController = 1;
            AnzahlZeichenProline = 24;
            break;
    
    // Display mit 4 linen  20 Zeichen (KS0073 Controller)
    case 3: DDRAM_Offset_line_1 = 0;
            DDRAM_Offset_line_2 = 32;
            DDRAM_Offset_line_3 = 64;
            DDRAM_Offset_line_4 = 96;
            AnzahlController = 1;
            AnzahlZeichenProline = 20;
            break;
    
    // Display mit 2 linen  16 Zeichen
    case 4: DDRAM_Offset_line_1 = 0;
            DDRAM_Offset_line_2 = 64;
            AnzahlController = 1;
            AnzahlZeichenProline = 16;
            break;
    }
  
  }

void LCD_WriteInstr(unsigned char LCD_Controller, unsigned char Instr)
  {
  unsigned char enable;
  
  switch (LCD_Controller)
    {
    case 1: enable = B4; // Ersten LCD_Controller auf Display aktivieren
            break;
    
    case 2: enable = B4; // Zweiten LCD_Controller auf Display aktivieren
            break;
    
    case 3: enable = B4; // Beide LCD_Controller auf Display aktivieren
            break;
    
    default: enable = B4;
    }
    
  // High Nibble senden
  POutReg = enable;            // Enable1, Enable2
  POutReg = enable+(Instr>>4); // Enable und HighNibble von Instr
  POutReg = Instr>>4;                 // HighNibble von Instr
  LCD_Wait(enable);   // 37us warten  // wahrscheinlich nicht notwendig...
    
  // Low Nibble senden
  Instr = Instr & (B0+B1+B2+B3);
  POutReg = enable;     // Enable1, Enable2
  POutReg = enable | Instr; // Enable und LowNibble von Instr
  POutReg = Instr;
  LCD_Wait(enable); // 37us warten
  }

void LCD_Clear(void)
  {
  LCD_WriteInstr(3, 1);
  }
  
void LCD_Control(unsigned char LCD_Controller, unsigned char disp_on, unsigned char cursor_on, unsigned char blink)
  {
  if (disp_on != 0)
    {
    disp_on = 4;
    }
  
  if (cursor_on != 0)
    {
    cursor_on = 2;
    }
    
  if (blink != 0)
    {
    blink = 1;
    }
    
  LCD_WriteInstr(LCD_Controller, B3 | disp_on | cursor_on | blink);
  }
  
char LCD_SetCursor(unsigned char line, unsigned char index)
  {
  unsigned char LCD_Controller = 1;
  unsigned char LCD_DDRAM_Adresse = 0;
    
  switch (line)
    {
    case 1: LCD_Controller = 1;
            LCD_DDRAM_Adresse = index-1 + DDRAM_Offset_line_1;
            break;
    
    case 2: LCD_Controller = 1;
	    LCD_DDRAM_Adresse = index-1 + DDRAM_Offset_line_2;
            break;
    
    case 3: if (AnzahlController == 1)
              {
              LCD_Controller = 1;
              }
            else
              {
              LCD_Controller = 2;
	      }
            LCD_DDRAM_Adresse = index-1 + DDRAM_Offset_line_3;
            break;
    
    case 4: if (AnzahlController == 1)
              {
              LCD_Controller = 1;
              }
            else
              {
              LCD_Controller = 2;
	      }
	    LCD_DDRAM_Adresse = index-1 + DDRAM_Offset_line_4;
            break;
    }
  
  // X-Position definieren ber DDRAM-Adresse
  LCD_WriteInstr(LCD_Controller, B7+LCD_DDRAM_Adresse);

  return(LCD_Controller);    
  }

void LCD_WriteChar(unsigned char LCD_Controller, unsigned char Data)
  {
  unsigned char RegisterSelect;
  unsigned char Enable;
  unsigned char Konvertiert;
  
  // Zeichenkonvertierung
  Konvertiert = Data;
  if (LCD_DisplayType != 3)
    {
    switch (Data)
      {
      case 228: Konvertiert = 225; // 
                break;

      case 246: Konvertiert = 239; // 
                break;

      case 252: Konvertiert = 245; // 
                break;

      case 196: Konvertiert = 225; // 
                break;

      case 214: Konvertiert = 239; // 
                break;

      case 220: Konvertiert = 245; // 
                break;
      }
    }
  else
    {
    switch (Data)
      {
      case 228: Konvertiert = 123; // 
                break;

      case 246: Konvertiert = 124; // 
                break;

      case 252: Konvertiert = 126; // 
                break;

      case 196: Konvertiert = 91;  // 
                break;

      case 214: Konvertiert = 92;  // 
                break;

      case 220: Konvertiert = 94;  // 
                break;
      
      case 91:  Konvertiert = 250; // [
                break;
      
      case 93:  Konvertiert = 252; // ]
                break;
      
      case 176: Konvertiert = 128; // 
                break;
      }
    }
    
  Data = Konvertiert;
  
  //if ((Data < 16) | (Data > 31)) // Nur druckbare Zeichen ausgeben
    { 
    RegisterSelect = B6; // DDRAM
  
    // Gewnschten LCD-Controller auf Display aktivieren	  
    switch (LCD_Controller)
      {
      case 1: Enable = B4;
              break;

      case 2: Enable = B4;
              break;

      case 3: Enable = B4;
              break;

      default: Enable = B4;
      }

    // High Nibble senden
    POutReg = Enable | RegisterSelect;
    POutReg = Enable | (Data>>4) | RegisterSelect;
    POutReg = (Data>>4) | RegisterSelect;
    LCD_Wait(Enable);

    // Low Nibble senden
    Data = Data & (B0+B1+B2+B3);
    POutReg = Enable | RegisterSelect;
    POutReg = Enable | Data | RegisterSelect;
    POutReg = Data | RegisterSelect;
    LCD_Wait(Enable);
    }
  }
  
void LCD_WriteStr(unsigned char line, unsigned char index, const unsigned char *str)
  {
  unsigned char LCD_Controller;
    
  LCD_Controller = LCD_SetCursor(line, index); // Cursor auf gewhlte Position setzen  
    
  while (*str != '\0')
    {
    LCD_WriteChar(LCD_Controller, *str); // Zeichen ausgeben
    str++;              
    }  
  }
  
void LCD_WriteInt(unsigned char line, unsigned char index, int zahl)
  {
  unsigned char ziffer;
  unsigned char flag;
  unsigned char LCD_Controller;
  unsigned char str_zahl;
  unsigned int maske;
  
  LCD_Controller = LCD_SetCursor(line, index); // Cursor auf gewhlte Position setzen
    
  if (zahl < 0)
    {
    LCD_WriteChar(LCD_Controller, '-');          // Bei negativer Zahl ein Minusindex voranstellen
    zahl = 32768 - (zahl & ~32768);              // Minus entfernen; absoluten Wert bernehmen 
    }
    
  if (zahl == 0)
    {
    LCD_WriteChar(LCD_Controller, '0');          // Wenn Zahl null ist, dies direkt ausgeben
    }
  else
    {
    maske = 10000;                               // Maske auf erste Ziffer im Integerwert setzen
    flag = 0;                                    // Nullunterdrckung am Anfang der Zahl
      
    while (maske > 0)
      {
      ziffer = zahl / maske;                     // Wert der Ziffer ermitteln
        
      if (ziffer != 0)
        {
        flag = 1;                                // Beim ersten Auftreten einer Zahl <> null, Zeichenunterdrckung aufheben
        }
        
      if (flag == 1)
        {
        str_zahl = '0' + ziffer;                 // ASCII Wert der Ziffer berechnen
        LCD_WriteChar(LCD_Controller, str_zahl); // Ziffer auf Display schreiben
        }
      zahl = zahl - (maske * ziffer);            // Ausgegebene Ziffer aus Zahl lschen
      maske = maske / 10;                        // Maske auf nchste Ziffer setzen
      }
    }
  }
  
void LCD_ClearLine(unsigned char line)
  {
  char LCD_Controller;
  char index;
    
  LCD_Controller = LCD_SetCursor(line, 1);
  for (index=1; index<=AnzahlZeichenProline; index++)
    {
    LCD_WriteChar(LCD_Controller, ' ');
    }  
  }
